=== modified file '.bzrignore'
--- .bzrignore	2005-10-19 11:34:00 +0000
+++ .bzrignore	2006-02-06 14:04:32 +0000
@@ -1,2 +1,3 @@
 bzrlib
 test
+.shelf

=== modified file 'NEWS'
--- NEWS	2005-11-29 07:12:26 +0000
+++ NEWS	2006-02-12 03:22:43 +0000
@@ -1,5 +1,24 @@
 NEWS
 
+Shelf version 0.8-rc:
+
+ * Shelves are now stored in .shelf, sigh. The bzr API for storing things under
+   .bzr is not rich enough at the moment to do all the things the shelf code
+   wants. Switching to .shelf also makes sense if I ever get around to making
+   shelf work on its own.
+
+ * There is a new command 'shelf', which has a bunch of subcommands for
+   operating on the shelf.
+   The current subcommands are:
+    list   (ls)           List the patches currently shelved.
+    delete (del) <patch>  Delete a patch from the shelf.
+
+ * Add .shelf to DEFAULT_IGNORE so it's ignored automatically.
+
+ * A few diffstat fixes, thanks to Dafydd Harries.
+
+ * Better error/status messages.
+
 Shelf version 0.7-rc:
 
  * Shelves are now stored in .bzr/x-shelf.

=== modified file '__init__.py'
--- __init__.py	2005-11-29 19:15:49 +0000
+++ __init__.py	2006-02-21 06:56:51 +0000
@@ -1,12 +1,15 @@
 #!/usr/bin/python
 """Shelf - temporarily set aside changes, then bring them back."""
 
-from sys import stderr
 import bzrlib.commands
 import bzrlib.branch
-from bzrlib.errors import BzrCommandError
+from errors import CommandError
 from bzrlib.option import Option
+from patchsource import BzrPatchSource
 from shelf import Shelf
+from bzrlib import DEFAULT_IGNORE
+
+DEFAULT_IGNORE.append('./.shelf')
 
 class cmd_shelve(bzrlib.commands.Command):
     """Temporarily set aside some changes to the current working tree.
@@ -29,32 +32,72 @@
     takes_args = ['file*']
     takes_options = [Option('pick'), 'message', 'revision']
     def run(self, pick=False, file_list=None, message=None, revision=None):
-        if file_list is not None and len(file_list) > 0:
-            branchdir = file_list[0]
-        else:
-            branchdir = '.'
-
         if revision is not None and revision:
             if len(revision) == 1:
                 revision = revision[0]
             else:
-                raise BzrCommandError("shelve only accepts a single revision "
+                raise CommandError("shelve only accepts a single revision "
                                   "parameter.")
 
-        s = Shelf(branchdir)
-        s.shelve(pick, message, revision, file_list)
+        source = BzrPatchSource(revision, file_list)
+        s = Shelf(source.branch.base)
+        s.shelve(source, pick, message)
         return 0
 
+class cmd_shelf(bzrlib.commands.Command):
+    """Perform various operations on your shelved patches.
+
+    Subcommands:
+        list   (ls)           List the patches on the current shelf.
+        delete (del) <patch>  Delete a patch from the current shelf.
+        switch       <shelf>  Switch to the named shelf, create it if necessary.
+        show         <patch>  Show the contents of the specified patch.
+    """
+    takes_args = ['subcommand', 'args*']
+
+    def run(self, subcommand, args_list):
+        import sys
+
+        source = BzrPatchSource()
+        s = Shelf(source.branch.base)
+
+        if subcommand == 'ls' or subcommand == 'list':
+            self.__check_no_args(args_list, "shelf list takes no arguments!")
+            s.list()
+        elif subcommand == 'delete' or subcommand == 'del':
+            self.__check_one_arg(args_list, "shelf delete takes one argument!")
+            s.delete(args_list[0])
+        elif subcommand == 'switch':
+            self.__check_one_arg(args_list, "shelf switch takes one argument!")
+            s = Shelf(source.branch.base, args_list[0])
+            s.make_default()
+        elif subcommand == 'show':
+            self.__check_one_arg(args_list, "shelf show takes one argument!")
+            s.display(args_list[0])
+        else:
+            print subcommand, args_list
+            print >>sys.stderr, "Unknown shelf subcommand '%s'" % subcommand
+
+    def __check_one_arg(self, args, msg):
+        if args is None or len(args) != 1:
+            raise BzrCommandError(msg)
+
+    def __check_no_args(self, args, msg):
+        if args is not None:
+            raise BzrCommandError(msg)
+
 class cmd_unshelve(bzrlib.commands.Command):
     """Reinstate the most recently shelved changes.
     See 'shelve' for more information.
     """
     takes_options = [Option('pick')]
     def run(self, pick=False):
-        s = Shelf('.')
-        s.unshelve(pick)
+        source = BzrPatchSource()
+        s = Shelf(source.branch.base)
+        s.unshelve(source, pick)
         return 0
 
+bzrlib.commands.register_command(cmd_shelf)
 bzrlib.commands.register_command(cmd_shelve)
 bzrlib.commands.register_command(cmd_unshelve)
 

=== added file 'errors.py'
--- errors.py	1970-01-01 00:00:00 +0000
+++ errors.py	2006-02-06 14:04:14 +0000
@@ -0,0 +1,5 @@
+try:
+    from bzrlib.errors import BzrCommandError as CommandError
+except ImportError:
+    class CommandError(Exception):
+        pass

=== modified file 'patches.py'
--- patches.py	2005-10-19 09:34:33 +0000
+++ patches.py	2006-02-11 03:27:01 +0000
@@ -302,6 +302,8 @@
     for line in iter_lines:
         if line.startswith('*** '):
             continue
+        if line.startswith('#'):
+            continue
         if line.startswith('==='):
             continue
         elif line.startswith('--- '):

=== added file 'patchsource.py'
--- patchsource.py	1970-01-01 00:00:00 +0000
+++ patchsource.py	2006-02-11 03:48:05 +0000
@@ -0,0 +1,51 @@
+import patches
+
+class PatchSource(object):
+    def __iter__(self):
+        def iterator(obj):
+            for p in obj.read():
+                yield p
+        return iterator(self)
+
+    def can_live_update(self):
+        return False
+
+    def readlines(self):
+        raise NotImplementedError()
+
+    def readhunks(self):
+        return patches.parse_patches(self.readlines())
+
+class FilePatchSource(PatchSource):
+    def __init__(self, filename):
+        self.filename = filename
+        PatchSource.__init__(self)
+
+    def readlines(self):
+        f = open(self.filename, 'r')
+        return f.readlines()
+
+class BzrPatchSource(PatchSource):
+    def __init__(self, revision=None, file_list=None):
+        from bzrlib.branch import Branch
+        self.revision = revision
+        self.file_list = file_list
+        if file_list is not None and len(file_list) > 0:
+            location = file_list[0]
+        else:
+            location = '.'
+        self.branch = Branch.open_containing(location)[0]
+
+        PatchSource.__init__(self)
+
+    def can_live_update(self):
+        return True
+
+    def readlines(self):
+        from StringIO import StringIO
+        from bzrlib.diff import show_diff
+        f = StringIO()
+        show_diff(self.branch, self.revision,
+                specific_files=self.file_list, output=f)
+        f.seek(0)
+        return f.readlines()

=== modified file 'shelf.py'
--- shelf.py	2005-12-04 20:51:27 +0000
+++ shelf.py	2006-02-21 05:05:27 +0000
@@ -1,137 +1,212 @@
 #!/usr/bin/python
 
-from patches import parse_patches
 import os
 import sys
-from bzrlib.commands import Command
-from bzrlib.branch import Branch
-from bzrlib.errors import BzrCommandError
+from errors import CommandError
 from hunk_selector import ShelveHunkSelector, UnshelveHunkSelector
 from diffstat import DiffStat
+from patchsource import PatchSource, FilePatchSource
 
 class Shelf(object):
-    def __init__(self, location, name='default'):
-        self.branch = Branch.open_containing(location)[0]
-        base = self.branch.controlfilename('x-shelf')
-        self.shelf_dir = os.path.join(base, name)
-
-        # FIXME surely there's an easier way to do this?
-        t = self.branch._transport
-        for dir in [base, self.shelf_dir]:
-            if not t.has(dir):
-                t.mkdir(dir)
-
-    def __path(self, idx):
-        return os.path.join(self.shelf_dir, '%.2d' % idx)
-
-    def next_shelf(self):
-        index = 0
-        while True:
-            name = self.__path(index)
-            if not os.path.exists(name):
-                return name
-            index += 1
-
-    def last_shelf(self):
-        shelves = os.listdir(self.shelf_dir)
-        indexes = [int(f) for f in shelves]
+    MESSAGE_PREFIX = "# Shelved patch: "
+
+    _paths = {
+        'base'          : '.shelf',
+        'shelves'       : '.shelf/shelves',
+        'current-shelf' : '.shelf/current-shelf',
+    }
+
+    def __init__(self, base, name=None):
+        self.base = base
+        self.__setup()
+
+        if name is None:
+            current = os.path.join(self.base, self._paths['current-shelf'])
+            name = open(current).read().strip()
+
+        assert '\n' not in name
+        self.name = name
+
+        self.dir = os.path.join(self.base, self._paths['shelves'], name)
+        if not os.path.isdir(self.dir):
+            os.mkdir(self.dir)
+
+    def __setup(self):
+        # Create required directories etc.
+        for dir in [self._paths['base'], self._paths['shelves']]:
+            dir = os.path.join(self.base, dir)
+            if not os.path.isdir(dir):
+                os.mkdir(dir)
+
+        current = os.path.join(self.base, self._paths['current-shelf'])
+        if not os.path.exists(current):
+            f = open(current, 'w')
+            f.write('default')
+            f.close()
+
+    def make_default(self):
+        f = open(os.path.join(self.base, self._paths['current-shelf']), 'w')
+        f.write(self.name)
+        f.close()
+        self.log("Default shelf is now '%s'\n" % self.name)
+
+    def log(self, msg):
+        sys.stderr.write(msg)
+
+    def delete(self, patch):
+        path = self.__path_from_user(patch)
+        os.remove(path)
+
+    def display(self, patch):
+        path = self.__path_from_user(patch)
+        sys.stdout.write(open(path).read())
+
+    def list(self):
+        self.log("Patches on shelf '%s':" % self.name)
+        indexes = self.__list()
+        if len(indexes) == 0:
+            self.log(' None\n')
+            return
+        self.log('\n')
+        for index in indexes:
+            msg = self.get_patch_message(self.__path(index))
+            if msg is None:
+                msg = "No message saved with patch."
+            self.log(' %.2d: %s\n' % (index, msg))
+
+    def __path_from_user(self, patch_id):
+        try:
+            patch_index = int(patch_id)
+        except TypeError:
+            raise CommandError("Invalid patch name '%s'" % patch_id)
+
+        path = self.__path(patch_index)
+
+        if not os.path.exists(path):
+            raise CommandError("Patch '%s' doesn't exist on shelf %s!" % \
+                        (patch_id, self.name))
+
+        return path
+
+    def __path(self, index):
+        return os.path.join(self.dir, '%.2d' % index)
+
+    def next_patch(self):
+        indexes = self.__list()
+
+        if len(indexes) == 0:
+            next = 0
+        else:
+            next = indexes[-1] + 1
+        return self.__path(next)
+
+    def __list(self):
+        patches = os.listdir(self.dir)
+        indexes = [int(f) for f in patches]
         indexes.sort()
+        return indexes
+
+    def last_patch(self):
+        indexes = self.__list()
 
         if len(indexes) == 0:
             return None
 
         return self.__path(indexes[-1])
 
-    def get_shelf_message(self, shelf):
-        prefix = "# shelf: "
-        if not shelf.startswith(prefix):
+    def get_patch_message(self, patch_path):
+        patch = open(patch_path, 'r').read()
+
+        if not patch.startswith(self.MESSAGE_PREFIX):
             return None
-        return shelf[len(prefix):shelf.index('\n')]
-
-    def unshelve(self, pick_hunks=False):
-        shelf = self.last_shelf()
-
-        if shelf is None:
-            raise BzrCommandError("No shelf found in branch '%s'" % \
-                self.branch.base)
-
-        patches = parse_patches(open(shelf, 'r').readlines())
+        return patch[len(self.MESSAGE_PREFIX):patch.index('\n')]
+
+    def __show_status(self, source):
+        if source.can_live_update():
+            diff_stat = str(DiffStat(source.readlines()))
+            if len(diff_stat) > 0:
+                self.log('Diff status is now:\n' + diff_stat + '\n')
+            else:
+                self.log('No changes left in working tree.\n')
+
+    def unshelve(self, patch_source, pick_hunks=False):
+        patch_name = self.last_patch()
+
+        if patch_name is None:
+            raise CommandError("No patch found on shelf %s" % self.name)
+
+        hunks = FilePatchSource(patch_name).readhunks()
         if pick_hunks:
-            to_unshelve, to_remain = UnshelveHunkSelector(patches).select()
+            to_unshelve, to_remain = UnshelveHunkSelector(hunks).select()
         else:
-            to_unshelve = patches
+            to_unshelve = hunks
             to_remain = []
 
         if len(to_unshelve) == 0:
-            raise BzrCommandError('Nothing to unshelve')
-
-        print >>sys.stderr, "Reapplying shelved patches",
-        message = self.get_shelf_message(open(shelf, 'r').read())
-        if message is not None:
-            print >>sys.stderr, ' "%s"' % message
-        else:
-            print >>sys.stderr, ""
-
-        pipe = os.popen('patch -d %s -s -p0' % self.branch.base, 'w')
-        for patch in to_unshelve:
-            pipe.write(str(patch))
+            raise CommandError('Nothing to unshelve')
+
+        message = self.get_patch_message(patch_name)
+        if message is None:
+            message = ""
+        self.log('Reapplying shelved patches "%s"\n' % message)
+
+        pipe = os.popen('patch -d %s -s -p0' % self.base, 'w')
+        for hunk in to_unshelve:
+            pipe.write(str(hunk))
         pipe.flush()
 
         if pipe.close() is not None:
-            raise BzrCommandError("Failed running patch!")
+            raise CommandError("Failed running patch!")
 
         if len(to_remain) == 0:
-            os.remove(shelf)
+            os.remove(patch_name)
         else:
-            f = open(shelf, 'w')
-            for patch in to_remain:
-                f.write(str(patch))
+            f = open(patch_name, 'w')
+            for hunk in to_remain:
+                f.write(str(hunk))
             f.close()
 
-        diff_stat = DiffStat(self.get_patches(None, None))
-        print 'Diff status is now:\n', diff_stat
-
-    def get_patches(self, revision, file_list):
-        from StringIO import StringIO
-        from bzrlib.diff import show_diff
-        out = StringIO()
-        show_diff(self.branch, revision, specific_files=file_list, output=out)
-        out.seek(0)
-        return out.readlines()
-
-    def shelve(self, pick_hunks=False, message=None, revision=None,
-             file_list=None):
-        patches = parse_patches(self.get_patches(revision, file_list))
+        self.__show_status(patch_source)
+
+    def shelve(self, patch_source, pick_hunks=False, message=None):
+        from datetime import datetime
+
+        hunks = patch_source.readhunks()
 
         if pick_hunks:
-            to_shelve = ShelveHunkSelector(patches).select()[0]
+            to_shelve = ShelveHunkSelector(hunks).select()[0]
         else:
-            to_shelve = patches
+            to_shelve = hunks
 
         if len(to_shelve) == 0:
-            raise BzrCommandError('Nothing to shelve')
-
-        shelf = self.next_shelf()
-        print >>sys.stderr, "Saving shelved patches to", shelf
-        shelf = open(shelf, 'a')
-        if message is not None:
-            assert '\n' not in message
-            shelf.write("# shelf: %s\n" % message)
-        for patch in to_shelve:
-            shelf.write(str(patch))
-
-        shelf.flush()
-        os.fsync(shelf.fileno())
-        shelf.close()
-
-        print >>sys.stderr, "Reverting shelved patches"
-        pipe = os.popen('patch -d %s -sR -p0' % self.branch.base, 'w')
-        for patch in to_shelve:
-            pipe.write(str(patch))
+            raise CommandError('Nothing to shelve')
+
+        if message is None:
+            timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
+            message = "Changes shelved on %s" % timestamp
+
+        patch_name = self.next_patch()
+        self.log('Shelving to %s/%s: "%s"\n' % \
+                (self.name, os.path.basename(patch_name), message))
+
+        patch = open(patch_name, 'a')
+
+        assert '\n' not in message
+        patch.write("%s%s\n" % (self.MESSAGE_PREFIX, message))
+
+        for hunk in to_shelve:
+            patch.write(str(hunk))
+
+        patch.flush()
+        os.fsync(patch.fileno())
+        patch.close()
+
+        pipe = os.popen('patch -d %s -sR -p0' % self.base, 'w')
+        for hunk in to_shelve:
+            pipe.write(str(hunk))
         pipe.flush()
 
         if pipe.close() is not None:
-            raise BzrCommandError("Failed running patch!")
+            raise CommandError("Failed running patch!")
 
-        diff_stat = DiffStat(self.get_patches(None, None))
-        print 'Diff status is now:\n', diff_stat
+        self.__show_status(patch_source)

=== modified file 'tests.py'
--- tests.py	2005-12-12 17:33:26 +0000
+++ tests.py	2006-02-21 05:09:26 +0000
@@ -1,8 +1,9 @@
 #!/usr/bin/python
 
 import bzrlib.tests
+import os.path
 
-class ShelfTests(bzrlib.tests.TestCaseInTempDir):
+class ShelfTests(bzrlib.tests.TestCaseWithTransport):
     ORIGINAL = '\n\nhello test world\n\n'
     MODIFIED = '\n\ngoodbye test world\n\n'
     DIFF_HEADER = "=== modified file 'test_file'\n"
@@ -31,13 +32,8 @@
         self.__test_loop(10)
 
     def __test_loop(self, count):
-        from bzrlib.branch import Branch
-        b = Branch.initialize('.')
-
-        # Create a test file and commit it
-        file('test_file', 'w').write(self.ORIGINAL)
-        b.working_tree().add('test_file')
-        b.working_tree().commit(message='add test_file')
+        tree = self.make_branch_and_tree('.')
+        self.__create_and_add_test_file(tree)
 
         while count > 0:
             count -= 1
@@ -59,7 +55,9 @@
             self.assertEqual(file('test_file').read(), self.ORIGINAL)
 
             # Check the shelf is right
-            shelf = b._transport.get('.bzr/x-shelf/default/00').read()
+            shelf = open(os.path.join(tree.branch.base,
+                        '.shelf/shelves/default/00')).read()
+            shelf = shelf[shelf.index('\n') + 1:] # skip the message
             self.assertEqual(shelf, self.DIFF_1)
 
             # Unshelve
@@ -74,32 +72,29 @@
 
     def test_shelf_nothing_to_shelve(self):
         import os.path
-        from bzrlib.branch import Branch
-        b = Branch.initialize('.')
-
-        # Create a test file and commit it
-        file('test_file', 'w').write(self.ORIGINAL)
-        b.working_tree().add('test_file')
-        b.working_tree().commit(message='add test_file')
+        tree = self.make_branch_and_tree('.')
+        self.__create_and_add_test_file(tree)
 
         # Shelve the changes
         self.run_bzr('shelve', retcode=3)
 
-        if b._transport.has('.bzr/x-shelf/default/00'):
+        if os.path.exists(os.path.join(tree.branch.base,
+                '.shelf/shelves/default/00')):
             self.fail("Shelf exists, but it shouldn't")
 
+    def __create_and_add_test_file(self, tree):
+        self.build_tree_contents([('test_file', self.ORIGINAL)])
+        tree.add('test_file')
+        tree.commit(message='add test_file')
+
     def test_shelf_with_revision(self):
-        from bzrlib.branch import Branch
-        b = Branch.initialize('.')
+        tree = self.make_branch_and_tree('.')
 
-        # Create a test file and commit it
-        file('test_file', 'w').write(self.ORIGINAL)
-        b.working_tree().add('test_file')
-        b.working_tree().commit(message='add test_file')
+        self.__create_and_add_test_file(tree)
 
         # Modify the test file and commit it
-        file('test_file', 'w').write(self.MODIFIED)
-        b.working_tree().commit(message='update test_file')
+        self.build_tree_contents([('test_file', self.MODIFIED)])
+        tree.commit(message='update test_file')
 
         # Shelve the changes
         self.run_bzr('shelve', '-r', '1', retcode=0)
@@ -118,8 +113,7 @@
         self.assertEqual(file('test_file').read(), self.MODIFIED)
 
     def test_shelf_with_two_revisions(self):
-        from bzrlib.branch import Branch
-        b = Branch.initialize('.')
+        tree = self.make_branch_and_tree('.')
 
         cmd = 'shelve -r 1..2'
         (stdout, stderr) = self.run_bzr_captured(cmd.split(), retcode=None)
@@ -145,3 +139,48 @@
         file('foo', 'wb').write('baz')
         self.run_bzr('shelve', 'foo', retcode=0)
 
+    def test_shelf_show_basic(self):
+        tree = self.make_branch_and_tree('.')
+        self.__create_and_add_test_file(tree)
+        self.__test_show(tree, '00')
+
+    def __test_show(self, tree, patch):
+        # Modify the test file
+        self.build_tree_contents([('test_file', 'patch %s\n' % patch)])
+
+        # Shelve the changes
+        self.run_bzr('shelve', retcode=0)
+
+        # Make sure there is no diff anymore
+        self.assertEqual(self.capture('diff', retcode=0), '')
+
+        # Check the shelf is right
+        shelf = open(os.path.join(tree.branch.base,
+                    '.shelf/shelves/default', patch)).read()
+        self.assertTrue('patch %s' % patch in shelf)
+
+        # Check the shown output is right
+        shown = self.capture('shelf show %s' % patch, retcode=0)
+        self.assertEqual(shown, shelf)
+
+    def test_shelf_show_multi(self):
+        tree = self.make_branch_and_tree('.')
+        self.__create_and_add_test_file(tree)
+        self.__test_show(tree, '00')
+        self.__test_show(tree, '01')
+        self.__test_show(tree, '02')
+
+        # Now check we can show a previously shelved patch
+        shelf = open(os.path.join(tree.branch.base,
+                    '.shelf/shelves/default/00')).read()
+        self.assertTrue('patch 00' in shelf)
+
+        # Check the shown output is right
+        shown = self.capture('shelf show 00', retcode=0)
+        self.assertEqual(shown, shelf)
+
+    def test_shelf_show_with_no_patch(self):
+        tree = self.make_branch_and_tree('.')
+        stderr = self.run_bzr_captured(['shelf', 'show', '00'], retcode=None)[1]
+        self.assertTrue("Patch '00' doesn't exist on shelf default!" in stderr)
+

